@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 # "screenkey" is distributed under GNU GPLv3+, WITHOUT ANY WARRANTY.
 # Copyright(c) 2010-2012: Pablo Seminario <pabluk@gmail.com>
 # Copyright(c) 2015-2020: wave++ "Yuri D'Elia" <wavexx@thregr.org>
@@ -18,13 +17,11 @@ import gi
 gi.require_version('Gtk', '3.0')
 gi.require_version('Pango', '1.0')
 
-from gi.repository import GLib
-GLib.threads_init()
-
-from gi.repository import Gtk, Gdk, GdkPixbuf, Pango, GObject
+from gi.repository import GLib, Gtk, Gdk, GdkPixbuf, Pango, GObject
 import cairo
 
 
+# Gtk shortcuts
 START = Gtk.Align.START
 CENTER = Gtk.Align.CENTER
 END = Gtk.Align.END
@@ -37,15 +34,20 @@ HORIZONTAL = Gtk.Orientation.HORIZONTAL
 VERTICAL = Gtk.Orientation.VERTICAL
 IF_VALID = Gtk.SpinButtonUpdatePolicy.IF_VALID
 
+
+BUTTONS_MIN_BLINK = 1/30        # Minimum persistence for any action (s)
+BUTTONS_REL_BRIGHT = 127        # Residual brightness after button release
+
+
+# SVG Data for mouse buttons
 BUTTONS_SVG = None
 
 def load_button_pixbufs(color):
     global BUTTONS_SVG
 
     if BUTTONS_SVG is None:
-        module_path = os.path.dirname(__file__)
-        image_path = os.path.join(module_path, 'images', 'mouse.svg')
-        with open(image_path, 'r') as svg_file:
+        image_path = os.path.join(MODULE_DIR, 'images', 'mouse.svg')
+        with open(image_path) as svg_file:
             BUTTONS_SVG = svg_file.readlines()
 
     if not isinstance(color, str):
@@ -80,22 +82,10 @@ def gi_module_available(module, version):
         return False
 
 
-class FixedSizeLabel(Gtk.Label):
-    def __init__(self):
-        super().__init__()
-
-    def do_get_preferred_height(self):
-        # Inhibit Label's default size negotiation so that it scales
-        # according to window's geometry unconditionally
-        return 0
-
-
 class Screenkey(Gtk.Window):
     STATE_FILE = os.path.join(GLib.get_user_config_dir(), 'screenkey.json')
 
     def __init__(self, logger, options, show_settings=False):
-        Gtk.Window.__init__(self, Gtk.WindowType.POPUP)
-
         self.logger = logger
         self.logger.debug("{} {}".format(APP_NAME, VERSION))
 
@@ -110,6 +100,7 @@ class Screenkey(Gtk.Window):
                             'ignore': [],
                             'position': 'bottom',
                             'persist': False,
+                            'window': False,
                             'font_desc': 'Sans Bold',
                             'font_size': 'medium',
                             'font_color': 'white',
@@ -141,6 +132,13 @@ class Screenkey(Gtk.Window):
                 if v is not None:
                     self.options[k] = v
 
+
+        if not self.options.window:
+            Gtk.Window.__init__(self, Gtk.WindowType.POPUP)
+        else:
+            self.options.persist = True
+            Gtk.Window.__init__(self, Gtk.WindowType.TOPLEVEL)
+
         self.set_keep_above(True)
         self.set_accept_focus(False)
         self.set_focus_on_map(False)
@@ -155,7 +153,7 @@ class Screenkey(Gtk.Window):
         self.box.show()
         self.add(self.box)
 
-        self.label = FixedSizeLabel()
+        self.label = Gtk.Label()
         self.label.set_ellipsize(Pango.EllipsizeMode.START)
         self.label.set_justify(Gtk.Justification.CENTER)
         self.label.show()
@@ -170,7 +168,7 @@ class Screenkey(Gtk.Window):
         self.connect("draw", self.on_draw)
 
         scr = self.get_screen()
-        scr.connect("size-changed", self.on_configure)
+        scr.connect("size-changed", self.on_screen_size_changed)
         scr.connect("monitors-changed", self.on_monitors_changed)
         self.set_active_monitor(self.options.screen)
 
@@ -212,10 +210,10 @@ class Screenkey(Gtk.Window):
         """Load stored options"""
         options = None
         try:
-            with open(self.STATE_FILE, 'r') as f:
+            with open(self.STATE_FILE) as f:
                 options = Options(json.load(f))
                 self.logger.debug("Options loaded.")
-        except IOError:
+        except OSError:
             self.logger.debug("file %s does not exists." % self.STATE_FILE)
         except ValueError:
             self.logger.debug("file %s is invalid." % self.STATE_FILE)
@@ -231,9 +229,9 @@ class Screenkey(Gtk.Window):
         """Store options"""
         try:
             with open(self.STATE_FILE, 'w') as f:
-                json.dump(options._store, f)
+                json.dump(options._store, f, indent=4)
                 self.logger.debug("Options saved.")
-        except IOError:
+        except OSError:
             self.logger.debug("Cannot open %s." % self.STATE_FILE)
 
 
@@ -246,7 +244,7 @@ class Screenkey(Gtk.Window):
         self.update_geometry()
 
 
-    def on_monitors_changed(self, *_):
+    def on_monitors_changed(self, monitor):
         self.set_active_monitor(self.monitor)
 
 
@@ -265,12 +263,15 @@ class Screenkey(Gtk.Window):
                 self.update_image_tag = None
 
 
+    def do_get_preferred_height(self):
+        return self.height
+
+
     def update_font(self):
-        window_width, window_height = self.get_size()
         text = self.label.get_text()
         lines = text.count('\n') + 1
-        self.font.set_absolute_size((50 * window_height // lines // 100) * 1000)
-        self.label.set_padding(window_width // 100, 0)
+        self.font.set_absolute_size((50 * self.height // lines // 100) * 1000)
+        self.label.set_padding(self.width // 100, 0)
         self.label.get_pango_context().set_font_description(self.font)
 
 
@@ -285,19 +286,14 @@ class Screenkey(Gtk.Window):
         for index, button_state in enumerate(self.button_states):
             if button_state is None:
                 continue
-            if button_state.pressed:
+            delta_time = (datetime.now() - button_state.stamp).total_seconds()
+            if button_state.pressed or delta_time < BUTTONS_MIN_BLINK:
                 alpha = 255
+            elif self.options.button_hide_duration > 0:
+                hide_time = delta_time / self.options.button_hide_duration
+                alpha = int(BUTTONS_REL_BRIGHT * (1 - min(1, hide_time)))
             else:
-                if self.options.button_hide_duration > 0:
-                    delta_time = (datetime.now() - button_state.stamp).total_seconds()
-                    hide_time = delta_time / self.options.button_hide_duration
-                else:
-                    hide_time = 1
-                if hide_time < 1:
-                    alpha = int(255 * (1 - hide_time))
-                else:
-                    self.button_states[index] = None
-                    continue
+                alpha = 0
 
             if not copied:
                 pixbuf = pixbuf.copy()
@@ -305,14 +301,16 @@ class Screenkey(Gtk.Window):
             self.button_pixbufs[button_state.btn].composite(
                 pixbuf, 0, 0, pixbuf.get_width(), pixbuf.get_height(),
                 0, 0, 1, 1,
-                GdkPixbuf.InterpType.NEAREST, alpha
-            )
+                GdkPixbuf.InterpType.NEAREST, alpha)
 
-        _, height = self.get_size()
-        scale = height / pixbuf.get_height()
+            if not button_state.pressed and \
+               delta_time >= self.options.button_hide_duration:
+                self.button_states[index] = None
+
+        scale = self.height / pixbuf.get_height()
         if scale != 1:
             width = int(pixbuf.get_width() * scale)
-            pixbuf = pixbuf.scale_simple(width, height, GdkPixbuf.InterpType.BILINEAR)
+            pixbuf = pixbuf.scale_simple(width, self.height, GdkPixbuf.InterpType.BILINEAR)
         self.img.set_from_pixbuf(pixbuf)
 
         if not copied:
@@ -341,10 +339,13 @@ class Screenkey(Gtk.Window):
         return False
 
 
-    def on_configure(self, *_):
+    def on_configure(self, event, data):
         # set event mask for click-through
         self.input_shape_combine_region(cairo.Region(cairo.RectangleInt(0, 0, 0, 0)))
 
+
+    def on_screen_size_changed(self, screen):
+        self.width, self.height = self.get_size()
         self.update_font()
         self.update_image()
 
@@ -400,10 +401,10 @@ class Screenkey(Gtk.Window):
 
 
     def move_resize(self, x, y, w, h):
+        self.width = w
+        self.height = h
         self.move(x, y)
         self.resize(w, h)
-        self.update_font()
-        self.update_image()
 
 
     def on_statusicon_popup(self, widget, button, timestamp, data=None):
@@ -412,10 +413,6 @@ class Screenkey(Gtk.Window):
             data.popup_at_pointer(None)
 
 
-    def show(self):
-        super(Screenkey, self).show()
-
-
     def on_labelmngr_error(self):
         msg = Gtk.MessageDialog(parent=self,
                                 type=Gtk.MessageType.ERROR,
@@ -436,8 +433,14 @@ class Screenkey(Gtk.Window):
             self.show()
         if self.timer_hide is not None:
             GObject.source_remove(self.timer_hide)
-        if self.options.timeout > 0 and not any(b and b.pressed for b in self.button_states):
-            self.timer_hide = GObject.timeout_add(self.options.timeout * 1000, self.on_timeout_main)
+            self.timer_hide = None
+        if self.options.timeout > 0:
+            # hide automatically if mouse mode is disabled. keep the
+            # window around otherwise as long as any of the visible keys
+            # (mouse or modifiers) is still held
+            if not self.options.mouse or \
+               not any(b and b.pressed for b in self.button_states):
+                self.timer_hide = GObject.timeout_add(self.options.timeout * 1000, self.on_timeout_main)
 
 
     def on_label_change(self, markup, synthetic):
@@ -453,6 +456,7 @@ class Screenkey(Gtk.Window):
         self.timed_show()
         if self.timer_min is not None:
             GObject.source_remove(self.timer_min)
+            self.timer_min = None
         if not synthetic:
             self.timer_min = GObject.timeout_add(self.options.recent_thr * 2000, self.on_timeout_min)
 
@@ -1027,6 +1031,56 @@ class Screenkey(Gtk.Window):
         self.about.show()
 
 
+    def start_lockscreen_detection(self):
+        from re import match
+        from threading import Thread
+        from dbus import SessionBus
+        from dbus.mainloop.glib import DBusGMainLoop
+
+        def filter_bus_message(bus, message):
+            message_member = message.get_member()
+
+            if not self.enabled or message_member != "ActiveChanged":
+                return
+
+            args_list = message.get_args_list()
+            if args_list[0]:
+                self.labelmngr.stop()
+                self.logger.debug("Lock Screen; Screenkey disabled.")
+            else:
+                self.restart_labelmanager()
+                self.logger.debug("Unlock Screen; Screenkey enabled.")
+
+        def lockscreen_detection_loop():
+            DBusGMainLoop(set_as_default=True)
+            session_bus = SessionBus()
+
+            signal_interface = None
+
+            for dbus_string in session_bus.list_names():
+                bus_name = str(dbus_string)
+                if match(r"org\.(\w+)\.ScreenSaver", bus_name):
+                    signal_interface = bus_name
+                    self.logger.debug(f"DBUS signal interface found: \"{signal_interface}\" ; password should not show when unlocking the screen.")
+                    break
+
+            if not signal_interface:
+                self.logger.debug("ScreenSaver DBUS signal interface not found; beware: password may show when unlocking the screen!")
+                del(session_bus)
+                DBusGMainLoop(set_as_default=False)
+                return
+
+            session_bus.add_match_string(f"type='signal',interface='{signal_interface}'")
+            session_bus.add_message_filter(filter_bus_message)
+            mainloop = GLib.MainLoop()
+            mainloop.run()
+
+        thread = Thread(target=lockscreen_detection_loop)
+        thread.daemon = True
+        thread.start()
+
+
     def run(self):
+        self.start_lockscreen_detection()
         Gtk.main()
         return self.exit_status
